/**
   This is a simple LoRa Gateway which routes recived traffic to a MQTT Broker
*/
/**
   This code (The Software) is supplied together with "MQTT Manager, Lora and Lora 'Poor Man' Gateway" Instructable and it is supplied to you
   by the Author in consideration of your agreement to the following terms, and your use or installation of The Software constitutes acceptance of these terms.
   If you do not agree with these terms, please do not use or install The Software.
   The Author grants you a personal, non-exclusive license, under authors copyrights in this original software, to use The Software.
   Except as expressly stated in this notice, no other rights or licenses, express or implied, are granted by the Author, including but not limited to any
   patent rights that may be infringed by your derivative works or by other works in which The Software may be incorporated.
   The Software is provided by the Author on an AS IS basis.  THE AUTHOR MAKES NO WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT
   LIMITATION THE IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE SOFTWARE OR ITS USE AND OPERATION
   ALONE OR IN COMBINATION WITH YOUR PRODUCTS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, INCIDENTAL OR CONSEQUENTIAL DAMAGES (INCLUDING,
   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE,
   REPRODUCTION AND MODIFICATION OF THE SOFTWARE HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT (INCLUDING NEGLIGENCE),
   STRICT LIABILITY OR OTHERWISE, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/

#include <WiFiClientSecure.h>
#include <MQTT.h>        // https://github.com/256dpi/arduino-mqtt
#include <SPI.h>
#include <LoRa.h>        // https://github.com/sandeepmistry/arduino-LoRa

#define DEBUG   //If you comment this line, the DPRINT & DPRINTLN lines are defined as blank.

#ifdef DEBUG
   #define DBEGIN(x)      Serial.begin(x) 
   #define DPRINT(...)    Serial.print(__VA_ARGS__)     
   #define DPRINTLN(...)  Serial.println(__VA_ARGS__)
#else
   #define DBEGIN(x)
   #define DPRINT(...)     
   #define DPRINTLN(...)
#endif

////

const char ssid[] = "[Your WiFi network SSID]";
const char pass[] = "[Your Password]";

// HiveMQ Broker - Secure connection
//
#define BROKER_ADDRESS "[MQTT Cluster Hostname]"
#define BROKER_PORT 8883
#define BROKER_USER "goofy"
#define BROKER_PASSWORD "Goofy01@"


/////

#define LORA_CS     5
#define LORA_RESET  17
#define LORA_IRQ    15
#define MONITOR_LED 32

byte msgCount = 0;            // count of outgoing messages
byte localAddress = 0xCC;     // address of this device

//

WiFiClientSecure net;
MQTTClient mqttClient;


//
bool routeIncomingMessage = false;
String sourceDeviceAddress;
String sourceDeviceProperty;
String sourceValue;

//
bool routeOutgoingMessage = false;
char destinationDeviceAddress[10];
char destinationDevicePropery[10];
char destinationValue[10];


void setup() {
  DBEGIN(115200);
  DPRINTLN("Start");

  //
  pinMode(MONITOR_LED,OUTPUT);
  digitalWrite(MONITOR_LED, HIGH);

  // WiFi Initializations

  WiFi.begin(ssid, pass);

  DPRINT("Connecting to WiFi ...");
  while (WiFi.status() != WL_CONNECTED) {
    DPRINT(".");
    delay(1000);
  }
  DPRINTLN(" connected!");
  net.setInsecure();

  // MQTT Initializations

  mqttClient.begin(BROKER_ADDRESS, BROKER_PORT, net);
  mqttClient.setOptions(60, false, 500);
  mqttClient.setWill("LoRa/gateway", "Disconnected", true, LWMQTT_QOS1);
  mqttClient.onMessage(mqttMessageReceived);

  connectToBroker();

  // Lora Initialization

  LoRa.setPins(LORA_CS, LORA_RESET, LORA_IRQ);
  
  // initialize radio at 868 MHz
  if (!LoRa.begin(868E6)) {
    DPRINTLN("LoRa init failed. Check your connections.");
    while (true)
      ;
  }

  DPRINTLN("LoRa init succeeded.");

  LoRa.onReceive(onLoraReceive);
  LoRa.receive();  
}

void registerDevice(String address, String device_input_property) {
  String topic = "LoRa/" + address + "/" + device_input_property;
  mqttClient.subscribe(topic);
}

void connectToBroker() {

  DPRINT("Connecting to MQTT Broker ...");

  while (!mqttClient.connect("ESP32", BROKER_USER, BROKER_PASSWORD)) {
    DPRINT("MQTT Client Disconnected! ");
    DPRINT(mqttClient.lastError()); DPRINTLN(" Return Code");
    DPRINTLN(mqttClient.returnCode());
    delay(1000);
  }

  DPRINTLN(" connected!");

  // Registration of each input property of each device
  registerDevice("ab", "S");

  // Last Will initialization
  mqttClient.publish("LoRa/gateway", "Ready", true, LWMQTT_QOS1);
}


void loop() {

  mqttClient.loop();

  if (routeOutgoingMessage) {
    routeOutgoingMessage = false;

    char buf[30];
    strcpy(&buf[0], &destinationDevicePropery[0]);
    strcat(&buf[0], "=");
    strcat(&buf[0], &destinationValue[0]);

    sendLoraMessage(strtol(destinationDeviceAddress, NULL, 16), buf);
    LoRa.receive();
  }

  if (routeIncomingMessage) {
    routeIncomingMessage = false;

    String topic = "LoRa/" + sourceDeviceAddress + "/" + sourceDeviceProperty;
    DPRINT("Sending Topic: ["); DPRINT(topic); DPRINT("] Payload: "); DPRINTLN(sourceValue);
    mqttClient.publish(topic, sourceValue, true, LWMQTT_QOS1);
  }
  delay(100);
  digitalWrite(MONITOR_LED, LOW);
  delay(500);
  digitalWrite(MONITOR_LED, HIGH);
}


void onLoraReceive(int packetSize) {
  if (packetSize == 0) return;          // if there's no packet, return

  // read packet header bytes:
  int recipient = LoRa.read();          // recipient address
  byte sender = LoRa.read();            // sender address
  byte incomingMsgId = LoRa.read();     // incoming msg ID
  byte incomingLength = LoRa.read();    // incoming msg length

  String incoming = "";                 // payload of packet

  while (LoRa.available()) {            
    incoming += (char)LoRa.read();
  }

  if (incomingLength != incoming.length()) {   // check length for error
    DPRINTLN("error: message length does not match length");
    return;
  }

  // if the recipient isn't this gateway
  if (recipient != localAddress) {
    DPRINTLN("This message is not for me.");
    return;
  }

  // if message is for this device, or broadcast, print details:
  DPRINTLN("Received from: 0x" + String(sender, HEX));
  DPRINTLN("Sent to: 0x" + String(recipient, HEX));
  DPRINTLN("Message ID: " + String(incomingMsgId));
  DPRINTLN("Message length: " + String(incomingLength));
  DPRINTLN("Message: " + incoming);
  DPRINTLN("RSSI: " + String(LoRa.packetRssi()));
  DPRINTLN();

  while(routeIncomingMessage == true)
    ;

  char buf[sizeof(incoming)];
  incoming.toCharArray(buf, sizeof(buf));

  sourceDeviceProperty = strtok(buf, "=");
  sourceDeviceAddress = String(sender, HEX);
  sourceValue = strtok(NULL, "=");

  routeIncomingMessage = true;
}


void mqttMessageReceived(String &topic, String &payload) {

  DPRINT("Received Topic: ["); DPRINT(topic); DPRINT("] Payload: "); DPRINTLN(payload);

  char buf[sizeof(topic)];
  topic.toCharArray(buf, sizeof(buf));

  char *p = strtok(buf, "/");

  p = strtok(0, "/");
  if (p == NULL)
    return;
  strcpy(&destinationDeviceAddress[0], p);

  p = strtok(0, "/");
  if (p == NULL)
    return;
  strcpy(&destinationDevicePropery[0], p);

  payload.toCharArray(destinationValue, sizeof(payload));

  routeOutgoingMessage = true;
}


void sendLoraMessage(byte device, String outgoing) {
  LoRa.beginPacket();                   // start packet
  LoRa.write(device);                   // add destination address
  LoRa.write(localAddress);             // add sender address
  LoRa.write(msgCount);                 // add message ID
  LoRa.write(outgoing.length());        // add payload length
  LoRa.print(outgoing);                 // add payload
  LoRa.endPacket();                     // finish packet and send it
  msgCount++;                           // increment message ID
}
